/**
 * Abstraction for threading to allow porting.
 * Currently using either pthread or C++ 11 threads.
 */

#ifndef PLS_THREAD_H
#define PLS_THREAD_H

#include <functional>
#include <pthread.h>

namespace pls {
    namespace internal {
        namespace base {
            using thread_entrypoint = void();

            class this_thread {
                template<typename Function, typename State>
                friend class thread;
                static pthread_key_t local_storage_key_;
                static bool local_storage_key_initialized_;

            public:
                static void yield() {
                    pthread_yield();
                }

                /**
                 * Retrieves the local state pointer.
                 *
                 * @tparam T The type of the state that is stored.
                 * @return The state pointer hold for this thread.
                 */
                template<typename T>
                static T* state() {
                    return reinterpret_cast<T*>(pthread_getspecific(local_storage_key_));
                }

                /**
                 * Stores a pointer to the thread local state object.
                 * The memory management for this has to be done by the user,
                 * we only keep the pointer.
                 *
                 * @tparam T The type of the state that is stored.
                 * @param state_pointer A pointer to the threads state object.
                 */
                template<typename T>
                static void set_state(T* state_pointer) {
                    pthread_setspecific(this_thread::local_storage_key_, (void*)state_pointer);
                }
            };

            template<typename Function, typename State>
            class thread {
                friend class this_thread;
                // Keep a copy of the function (lambda) in this object to make sure it is valid when called!
                Function function_;
                State* state_pointer_;

                // Keep handle to native implementation
                pthread_t pthread_thread_;

                static void* start_pthread_internal(void* thread_pointer) {
                    auto my_thread = reinterpret_cast<thread*>(thread_pointer);
                    this_thread::set_state(my_thread->state_pointer_);
                    my_thread->function_();
                    pthread_exit(nullptr);
                }

            public:
                explicit thread(const Function& function, State* state_pointer):
                    function_{function},
                    state_pointer_{state_pointer},
                    pthread_thread_{} {

                    if (!this_thread::local_storage_key_initialized_) {
                        pthread_key_create(&this_thread::local_storage_key_, nullptr);
                        this_thread::local_storage_key_initialized_ = true;
                    }

                    pthread_create(&pthread_thread_, nullptr, start_pthread_internal, (void *)(this));
                }
            public:
                void join() {
                    pthread_join(pthread_thread_, nullptr);
                }

                // make object move only
                thread(thread&&) noexcept = default;
                thread& operator=(thread&&) noexcept = default;

                thread(const thread&) = delete;
                thread& operator=(const thread&) = delete;
            };

            template<typename Function, typename State>
            thread<Function, State> start_thread(const Function& function, State* state_pointer) {
                return thread<Function, State>(function, state_pointer);
            }

            template<typename Function>
            thread<Function, void> start_thread(const Function& function) {
                return thread<Function, void>(function, nullptr);
            }
        }
    }

}

#endif //PLS_THREAD_H
